from inspect import isclass

from inflection import underscore

from microcosm_pubsub.conventions.lifecycle import LifecycleChange


def name_for(obj):
    """
    Get a name for something.

    Allows overriding of default names using the `__alias__` attribute.

    """
    if isinstance(obj, str):
        return underscore(obj)

    cls = obj if isclass(obj) else obj.__class__

    if hasattr(cls, "__alias__"):
        return underscore(cls.__alias__)
    else:
        return underscore(cls.__name__)


def make_media_type(resource, lifecycle_change=None, organization=None, public=False):
    """
    Generate a canonical media type for a message that announces a lifecycle change to a resource.

    :param resource: the resource (name)
    :param lifecycle_change: the change that happened to the resource
    :param organization: the vendor organization to which this resource belongs
    :param public: whether the resource is meant for public/external consumption

    """
    if lifecycle_change is None:
        lifecycle_change = LifecycleChange.Created
    if organization is None:
        organization = "globality"

    return "application/vnd.{}.{}.{}.{}".format(
        # use a vendor specific media type
        organization,
        # differentiate messages sent over a pubsub channel from other messages
        "pubsub" if public else "pubsub._",
        # qualify the message with the kind of lifecycle change
        str(lifecycle_change),
        # specify the resource name
        name_for(resource),
    )



EXTRACTS = "_extracts"
BINDS = "_binds"


def extracts(*extract):
    """
    On resolving this function with get_from_context decorator -
    updates the context based on the function return value.

    Note: The decorator just marks the function and don't change the return value.
    :param *args: key/s of the return value

    """
    def decorate(func):
        setattr(func, EXTRACTS, extract)
        return func
    return decorate


def binds(**binds):
    """
    On resolving this function with temporarily_replace_context_keys decorator -
    temporarily updates the context keys.

    Note: The decorator just marks the function.
    :param **binds: (old_context_key, new_context_key) - context keys to rename

    """
    def decorate(func):
        setattr(func, BINDS, binds)
        return func
    return decorate



"""
Microcosm Pubsub Models

"""


class SNSIntrospection(object):
    def __init__(
        self,
        media_type,
        route,
        call_module,
        call_function,
        count,
    ):
        self.media_type = media_type
        self.route = route
        self.call_module = call_module
        self.call_function = call_function
        self.count = count



class ArgumentExtractor:
    def __init__(self, name, key, key_property=None):
        self.name = name
        self.key = key
        self.key_property = key_property

    def __str__(self):
        return f"extract_{self.name}"

    def __call__(self, context):
        obj = context[self.key]
        if self.key_property is None:
            value = obj
        elif hasattr(obj, self.key_property):
            value = getattr(obj, self.key_property)
        else:
            value = obj[self.key_property]

        context[self.name] = value
        return value


class WhenStatement:
    def __init__(self, key):
        self.key = key
        self._then = None
        self._otherwise = None

    def then(self, chain):
        self._then = chain
        return self

    def otherwise(self, chain):
        self._otherwise = chain
        return self

    def __str__(self):
        return f"when_{self.key}"

    def __call__(self, context):
        if context[self.key]:
            if self._then:
                return self._then(context)
        elif self._otherwise:
            return self._otherwise(context)


class SwitchStatement:
    def __init__(self, key):
        self.key = key
        self._otherwise = None
        self._cases = dict()

    def __str__(self):
        return f"switch_{self.key}"

    def case(self, key, chain):
        self._cases[key] = chain
        return self

    def otherwise(self, chain):
        self._otherwise = chain
        return self

    def __call__(self, context):
        key = context[self.key]
        action = self._cases.get(key, self._otherwise)
        if action:
            return action(context)


class TryStatement:
    def __init__(self, chain):
        self.chain = chain
        self._cases = dict()
        self._otherwise = None

    def catch(self, key, chain):
        self._cases[key] = chain
        return self

    def otherwise(self, chain):
        self._otherwise = chain
        return self

    def __call__(self, context):
        try:
            res = self.chain(context)
        except Exception as excption:
            handle = self._cases.get(type(excption))
            if not handle:
                raise excption
            return handle(context)
        else:
            if self._otherwise:
                return self._otherwise(context)
            return res


def extract(name, key, key_property=None):
    """
    Extract an argument from a context to anotherwise context key

    :param name: new context key
    :param from_: old context key
    :param property_: propery of the context key

    """
    return ArgumentExtractor(name, key, key_property)


def when(key):
    """
    Run one of two chains - based on a condition

    Example: when("arg")
        .then(Chain(...))
        .otherwise(Chain(...))

    """

    return WhenStatement(key)


def switch(key):
    """
    Run one of number of chains chains - based on a key value
    Example: when("arg")
        .case(100, Chain(...))
        .case(200, Chain(...))
        .otherwise(Chain(...))

    """
    return SwitchStatement(key)


def try_chain(chain):
    """
    Run one of two chains - based on a condition

    Example: try_chain(Chain(...))
        .catch(ValueError, Chain(...))
        .catch(KeyError, Chain(...))
        .otherwise(Chain(...))

    """
    return TryStatement(chain)



from microcosm_pubsub.chain.context import SafeContext
from microcosm_pubsub.chain.context_decorators import (
    get_from_context,
    save_to_context,
    save_to_context_by_func_name,
    temporarily_replace_context_keys,
)


class Chain:
    """
    Structure that contains callable functions that shares the same context.

    """
    def __init__(self, *args):
        """
        Build a chain that can be resolved multiple times
        :param *args: callable functions, Chain statemnts or other Chains

        """
        self.links = list(args)

    @property
    def context_decorators(self):
        """
        Decorators to apply to the chain links

        """
        return [
            # Order matter - get_from_context should be first
            get_from_context,
            temporarily_replace_context_keys,
            save_to_context,
            save_to_context_by_func_name,
        ]

    @property
    def new_context_type(self):
        """
        Context type to create

        """
        return SafeContext

    def __call__(self, context=None, **kwargs):
        """
        Resolve the chain and return the last chain function result
        :param context: use existing context instead of creating a new one
        :param **kwargs: initialize the context with some values

        """
        context = context or self.new_context_type()
        context.update(kwargs)

        # Allow self reference
        if "context" not in context:
            context.update(context=context)
        elif context["context"] is not context:
            raise ValueError("context should be set as context")

        for link in self.links:
            res = self.apply_decorators(context, link)()
        return res

    def apply_decorators(self, context, link):
        decorated_link = link
        for decorator in self.context_decorators:
            decorated_link = decorator(context, decorated_link)
        return decorated_link



from collections import MutableMapping


class SafeContext(MutableMapping):
    """
    Dictionary that raises on overwriting keys.

    """

    def __init__(self, *args, **kwargs):
        self.store = dict(*args, **kwargs)

    def __getitem__(self, key):
        return self.store[key]

    def __setitem__(self, key, value):
        if key in self.store:
            raise ValueError(f"Key '{key}' already set")
        self.store[key] = value

    def __delitem__(self, key):
        del self.store[key]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def __getattr__(self, key):
        return self[key]



"""
Publish Info convention.

"""
from microcosm_flask.conventions.base import EndpointDefinition
from microcosm_flask.conventions.crud import configure_crud
from microcosm_flask.operations import Operation
from microcosm_flask.namespaces import Namespace
from microcosm_pubsub.conventions.publish_info.resources import (
    PublishInfoSchema,
    Schema,
)


def configure_publish_info(graph):
    ns = Namespace(
        subject="introspection/publish_info",
    )

    def search(**kwargs):
        publish_info = list(graph.sns_producer.get_publish_info())
        return publish_info, len(publish_info)

    mappings = {
        Operation.Search: EndpointDefinition(
            func=search,
            request_schema=Schema(),
            response_schema=PublishInfoSchema(),
        ),
    }

    configure_crud(graph, ns, mappings)
    return ns



"""
Consume Daemon main.

"""
from microcosm.loaders import load_each, load_from_dict
from microcosm_daemon.api import SleepNow
from microcosm_daemon.daemon import Daemon

from microcosm_pubsub.envelope import SQSEnvelope, NaiveSQSEnvelope
from microcosm_pubsub.consumer import STDIN


class ConsumerDaemon(Daemon):

    def __init__(self):
        super().__init__()
        self.bound_handlers = None

    def make_arg_parser(self):
        parser = super().make_arg_parser()
        parser.add_argument("--stdin", action="store_true")
        parser.add_argument("--sqs-queue-url")
        parser.add_argument("--envelope", choices=[
            envelope_cls.__name__
            for envelope_cls in SQSEnvelope.__subclasses__()
        ])
        return parser

    def create_object_graph_components(self, graph):
        super().create_object_graph_components(graph)
        self.bound_handlers = graph.sqs_message_handler_registry.compute_bound_handlers(
            self.components,
        )

    def run_state_machine(self):
        for media_type, handler in self.bound_handlers.items():
            self.graph.logger.info("Handling: {} with handler: {}".format(
                media_type,
                handler.__name__,
            ))

        super().run_state_machine()

    @property
    def defaults(self):
        config = dict()

        if self.args.stdin:
            config.update(
                sqs_envelope=dict(
                    strategy_name=NaiveSQSEnvelope.__name__,
                ),
                sqs_consumer=dict(
                    sqs_queue_url=STDIN,
                ),
            )

        if self.args.sqs_queue_url:
            config.update(
                sqs_consumer=dict(
                    sqs_queue_url=self.args.sqs_queue_url,
                ),
            )

        if self.args.envelope:
            config.update(
                sqs_envelope=dict(
                    strategy_name=self.args.envelope,
                ),
            )

        return config

    @property
    def components(self):
        return super().components + [
            "opaque",
            "pubsub_message_schema_registry",
            "sqs_message_handler_registry",
            "sqs_consumer",
            "sqs_message_dispatcher",
        ]

    def __call__(self, graph):
        """
        Implement daemon by sinking messages from the consumer to a dispatcher function.

        """
        result = graph.sqs_message_dispatcher.handle_batch(self.bound_handlers)
        if not result.message_count:
            raise SleepNow()

    @classmethod
    def create_for_testing(cls, loader=None, **kwargs):
        mock_config = load_from_dict(
            sns_producer=dict(
                mock_sns=False,
            ),
        )

        if loader is None:
            loader = mock_config
        else:
            loader = load_each(loader, mock_config)

        return super().create_for_testing(
            sqs_queue_url="queue",
            loader=loader,
            envelope=None,
            stdin=False,
            **kwargs
        )



from microcosm_pubsub.chain.chain import Chain  # noqa: F401
from microcosm_pubsub.chain.decorators import binds, extracts  # noqa: F401
from microcosm_pubsub.chain.statements import extract, when, switch, try_chain  # noqa: F401



"""
Publish Info resources.

"""
from marshmallow import fields, Schema


class PublishInfoSchema(Schema):
    mediaType = fields.String(attribute="media_type")
    route = fields.String()
    callModule = fields.String(attribute="call_module")
    callFunction = fields.String(attribute="call_function")
    count = fields.Integer()
